
/*
Author: Sean D. Spencer 
    (sean_don4@yahoo.com)
Date: 1/7/2006
*/

#include <Python.h>
#include <alsa/asoundlib.h>

#define DEFAULT_NAME  "mymidi"

static PyObject *AlsaError; /* locally raised exception */

#define RAISE(message) \
    { PyErr_SetString(AlsaError, message); return NULL; }

static snd_seq_t *gp_seq_handle = NULL;
static int g_client, g_port;
static int g_seq_client = 128; /* Default to 128 */
static int g_seq_port = 0;
static snd_seq_event_t g_ev;

static PyObject *seq_open(PyObject *self, PyObject *args) 
{
    unsigned int caps;

    if(! PyArg_ParseTuple(args, "")) /* no args */
	return NULL;

    /* sequencer opening */
    if (snd_seq_open(&gp_seq_handle, "hw", SND_SEQ_OPEN_OUTPUT, 0))
    {
	RAISE("Error opening ALSA sequencer.")
    }

    /* our client id */
    g_client = snd_seq_client_id(gp_seq_handle);

    /* set client info */
    snd_seq_set_client_name(gp_seq_handle, DEFAULT_NAME);

    /* create port */
    caps = SND_SEQ_PORT_CAP_READ;
    if (g_seq_client == SND_SEQ_ADDRESS_SUBSCRIBERS)
	caps |= SND_SEQ_PORT_CAP_SUBS_READ;

    g_port = snd_seq_create_simple_port(gp_seq_handle, DEFAULT_NAME,
	caps,SND_SEQ_PORT_TYPE_MIDI_GENERIC | SND_SEQ_PORT_TYPE_APPLICATION);

    if (g_port < 0)
    {
	RAISE("Unable to create port.")
	snd_seq_close(gp_seq_handle);
    }

    /* subscribe to MIDI port */
    if (g_seq_client != SND_SEQ_ADDRESS_SUBSCRIBERS)
    {
	if (snd_seq_connect_to(gp_seq_handle, g_port, g_seq_client, g_seq_port) < 0)
	{
	    RAISE("Unable to subscribe to MIDI port.")
	    snd_seq_close(gp_seq_handle);
	}
    }

    Py_INCREF(Py_None); /* reference count for Python garbage collection */
    return Py_None;	/* None: no errors */
}

static PyObject *seq_set_client(PyObject *self, PyObject *args)
{
    int client;

    if (! PyArg_ParseTuple(args, "i", &client))
	return NULL;	/* NULL triggers exception */

    g_seq_client = client;

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *seq_get_client(PyObject *self, PyObject *args)
{
    if(! PyArg_ParseTuple(args, "")) /* no args */
	return NULL;

    return Py_BuildValue("i", g_seq_client);
}

static PyObject *seq_close(PyObject *self, PyObject *args) 
{
    if(! PyArg_ParseTuple(args, "")) /* no args */
	return NULL;

    snd_seq_close(gp_seq_handle);

    Py_INCREF(Py_None);
    return Py_None;
}

void seq_send_event()
{
    snd_seq_ev_set_direct(&g_ev);
    snd_seq_ev_set_source(&g_ev, g_port);
    snd_seq_ev_set_dest(&g_ev, g_seq_client, g_seq_port);
    snd_seq_event_output(gp_seq_handle, &g_ev);
    snd_seq_drain_output(gp_seq_handle);
}

static PyObject *seq_program_change(PyObject *self, PyObject *args)
{
    int chan_no, preset;

    if (! PyArg_ParseTuple(args, "ii", &chan_no, &preset))
	return NULL;	/* NULL triggers exception */

    snd_seq_ev_set_pgmchange(&g_ev, chan_no, preset);
    seq_send_event();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *seq_control_change(PyObject *self, PyObject *args)
{
    int chan_no, type, value;

    if (! PyArg_ParseTuple(args, "iii", &chan_no, &type, &value))
	return NULL;	/* NULL triggers exception */

    snd_seq_ev_set_controller(&g_ev, chan_no, type, value);
    seq_send_event();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *seq_noteon(PyObject *self, PyObject *args)
{
    int chan, note, vel;

    if (! PyArg_ParseTuple(args, "iii", &chan, &note, &vel))
	return NULL;	/* NULL triggers exception */

    snd_seq_ev_set_noteon(&g_ev, chan, note, vel);
    seq_send_event();

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *seq_noteoff(PyObject *self, PyObject *args)
{
    int chan, note, vel;
    
    if (! PyArg_ParseTuple(args, "iii", &chan, &note, &vel))
	return NULL;	/* NULL triggers exception */

    snd_seq_ev_set_noteoff(&g_ev, chan, note, vel);
    seq_send_event();

    Py_INCREF(Py_None);
    return Py_None;
}

static struct PyMethodDef alsa_methods[] =
{
    {"seq_open", seq_open, 1},
    {"seq_set_client", seq_set_client, 1},
    {"seq_get_client", seq_get_client, 1},
    {"seq_close", seq_close, 1},
    {"seq_program_change", seq_program_change, 1},
    {"seq_control_change", seq_control_change, 1},
    {"seq_noteon", seq_noteon, 1},
    {"seq_noteoff", seq_noteoff, 1},
    {NULL, NULL}
};

void initalsa()
{
    PyObject *m;

    m = Py_InitModule("alsa", alsa_methods);

    AlsaError = PyErr_NewException("alsa.error", NULL, NULL);
    Py_INCREF(AlsaError);
    PyModule_AddObject(m, "error", AlsaError);

    if (PyErr_Occurred())
	Py_FatalError("Unable to initialize module alsa.");
}

